
# All packages used in this script
import os
import uuid
import shutil
import subprocess
import hashlib
import random
import time
from typing import List, Tuple, Optional
from typing import Optional, Tuple, Dict
import torch
from tqdm import tqdm
from esm.models.esm3 import ESM3
from esm.sdk.api import ESMProtein, GenerationConfig
from .guided_generation import ESM3GuidedDecoding, GuidedDecodingScoringFunction
from .scoring_utils import FoldXScorer, parse_pdb_chain_sequence_with_mapping, foldx_repair_pdb, plot_ddg_history
from multiprocessing import Pool, cpu_count, Manager
from functools import partial
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from multiprocessing import set_start_method
import warnings
import argparse


# --- Main Settings ---
# This ensures multiprtocessing works correctly in different environments

try:
    set_start_method('spawn')
except RuntimeError:
    pass
warnings.simplefilter(action='ignore', category=FutureWarning)

# --- Path and Execution Settings ---
FOLDX_EXEC     = "/pscratch/sd/a/ananda/ESM3-Guided-Generation-Based-Protein-Engineering/foldx/foldx_20251231"
FOLDX_WORKDIR  = "/pscratch/sd/a/ananda/ESM3-Guided-Generation-Based-Protein-Engineering/foldx"

# --- Protein and Masking Settings to test for individual protein without batch job ---

# PDB_FILENAME   = "2lwy.pdb"
# CHAIN_ID       = "A"
# MASKING_PERCENTAGE = 0.4


# --- Generation Step Settings ---

# NUM_DECODING_STEPS = 2
# NUM_SAMPLES_PER_STEP = 2
# NUM_WORKERS = 2 

# NUM_DECODING_STEPS = 32
# NUM_SAMPLES_PER_STEP = 20
# NUM_WORKERS = 20 

# --- FoldX Settings ---

NUMBER_OF_RUNS = 1
TIMEOUT_SEC    = 1800
CLEANUP_TMP    = True 
CACHE_DIR      = os.path.join(FOLDX_WORKDIR, "foldx_cache")
VERBOSE_FOLDX  = False 

USE_STARTING_MUTATIONS = False
STARTING_MUTATIONS = { 10: "F", 25: "Y", 30: "W", 41: "I", 55: "R" }


def main():
    parser = argparse.ArgumentParser(description="Run ESM3 Guided Generation with FoldX.")
    parser.add_argument(
        "--pdb_filename",
        type=str,
        required=True,
        help="The name of the PDB file located in the FOLDX_WORKDIR (e.g., '2lwy.pdb')."
    )
    parser.add_argument(
        "--chain_id",
        type=str,
        required=True,
        help="The single-letter chain ID to extract and redesign (e.g., 'A')."
    )
    parser.add_argument(
        "--masking_percentage",
        type=float,
        default=0.40,
        help="Percentage of non-frozen residues to mask (e.g., 0.4 for 40%)."
    )

    parser.add_argument(
        "--num_decoding_steps", type=int, default=32, help="Number of generation steps."
    )
    parser.add_argument(
        "--num_samples_per_step", type=int, default=20, help="Number of candidates to generate per step."
    )
    parser.add_argument(
        "--num_workers", type=int, default=20, help="Number of parallel workers for FoldX scoring."
    )

    
    args = parser.parse_args()

    PDB_FILENAME = args.pdb_filename
    CHAIN_ID = args.chain_id
    MASKING_PERCENTAGE = args.masking_percentage
    NUM_DECODING_STEPS = args.num_decoding_steps
    NUM_SAMPLES_PER_STEP = args.num_samples_per_step
    NUM_WORKERS = args.num_workers


    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    # --- 1. Initial Setup ---
    os.makedirs(FOLDX_WORKDIR, exist_ok=True); os.makedirs(CACHE_DIR, exist_ok=True)
    assert os.path.isfile(os.path.join(FOLDX_WORKDIR, "rotabase.txt")), "rotabase.txt missing"
    REPAIRED_PDB_PATH = os.path.join(FOLDX_WORKDIR, f"{os.path.splitext(PDB_FILENAME)[0]}_Repair.pdb")
    if not os.path.isfile(REPAIRED_PDB_PATH): REPAIRED_PDB_PATH = foldx_repair_pdb(FOLDX_EXEC, FOLDX_WORKDIR, PDB_FILENAME,TIMEOUT_SEC)
    WT_SEQ, SEQ_TO_PDB_MAP = parse_pdb_chain_sequence_with_mapping(REPAIRED_PDB_PATH, CHAIN_ID)

     # --- 2. Create the Masked Starting Sequence ---
    if USE_STARTING_MUTATIONS:
        print("[INFO] Applying starting mutations and freezing those positions.")
        mutant_seq_list = list(WT_SEQ)
        for pos, aa in STARTING_MUTATIONS.items():
            mutant_seq_list[pos - 1] = aa
        base_sequence = "".join(mutant_seq_list)
        
        frozen_indices_0based = {i - 1 for i in STARTING_MUTATIONS.keys()}
        maskable_indices_0based = list(set(range(len(base_sequence))) - frozen_indices_0based)

    else:
        print("[INFO] Using wild-type sequence as the base for masking.")
        base_sequence = WT_SEQ
        maskable_indices_0based = list(range(len(base_sequence)))

    
    num_to_mask = int(len(maskable_indices_0based) * MASKING_PERCENTAGE)
    indices_to_mask_0based = random.sample(maskable_indices_0based, num_to_mask)
    
    # refinement_template_list = list(starting_mutant_seq)
    refinement_template_list = list(base_sequence)
    
    for i in indices_to_mask_0based: 
        refinement_template_list[i] = '_'
    refinement_start_seq = "".join(refinement_template_list)

    
    # --- Setup the Log File ---
    timestamp = time.strftime("%Y%m%d_%H%M%S")
    log_filename = f"generation_log_{timestamp}.txt"
    log_dir = "../../logs"
    os.makedirs(log_dir, exist_ok=True)
    log_filepath = os.path.join(log_dir, log_filename)
    print(f"[INFO] All generated sequences will be saved to: {log_filepath}")

    header_text = (
        f"{'='*65}\n"
        f"ESM3-FoldX Guided Generation Log\n"
        f"Run Timestamp: {timestamp}\n"
        f"{'='*65}\n\n"
        f"--- RUN PARAMETERS ---\n"
        f"PDB File: {PDB_FILENAME}\n"
        f"Masking Percentage: {MASKING_PERCENTAGE * 100:.1f}%\n"
        f"Decoding Steps: {NUM_DECODING_STEPS}\n"
        f"Samples per Step: {NUM_SAMPLES_PER_STEP}\n\n"
        f"--- SEQUENCE SETUP ---\n"
        f"Wild-Type Sequence Length: {len(WT_SEQ)}\n"
        f"Original Wild-Type Sequence:\n{WT_SEQ}\n\n"
        f"Masked Template for Generation ({len(indices_to_mask_0based)} masks):\n{refinement_start_seq}\n"
        f"{'='*65}\n"
)
    with open(log_filepath, 'w') as f:
        f.write(header_text)
    print(header_text)
    
    # --- 3. Initialize Model and Run Guided Generation ---
    device = "cuda" if torch.cuda.is_available() else "cpu"
    print(f"\nLoading ESM3 model to {device}...")
    model = ESM3.from_pretrained().to(device).float()

    scorer_kwargs = {
        'foldx_exec': FOLDX_EXEC,
        'foldx_workdir': FOLDX_WORKDIR,
        'cache_dir': CACHE_DIR, 
        'number_of_runs': 1,
        'timeout_sec': 1800,
        'cleanup_tmp': True,
        'verbose_foldx': False,
}
    scoring_function = FoldXScorer(WT_SEQ, CHAIN_ID, SEQ_TO_PDB_MAP, REPAIRED_PDB_PATH,**scorer_kwargs)
    guided_decoding = ESM3GuidedDecoding(client=model, scoring_function=scoring_function)

    print(f"\n--- Running PARALLEL Guided Generation from the MASKED template ---")
    starting_protein = ESMProtein(sequence=refinement_start_seq)
    
    # Call the built-in function, which now supports parallelism and logging
    generated_protein, all_scores, best_overall_score, best_overall_step = guided_decoding.guided_generate(
        protein=starting_protein,
        num_decoding_steps=NUM_DECODING_STEPS,
        num_samples_per_step=NUM_SAMPLES_PER_STEP,
        track="sequence",
        num_workers=NUM_WORKERS,
        log_file_path=log_filepath
    )
    
    # --- 4. Display and Save Final Result ---
    if generated_protein:
        # best_last_step_score = max(s for s in all_scores.get(NUM_DECODING_STEPS, []) if s is not None and s != float('-inf'))

        actual_last_step = 0
        if all_scores:
            actual_last_step = max(all_scores.keys())

        # Now, get the best score from that actual last step
        best_last_step_score = -float('inf')
        if actual_last_step > 0:
            valid_scores_last_step = [s for s in all_scores.get(actual_last_step, []) if s is not None and s != float('-inf')]
            if valid_scores_last_step:
                best_last_step_score = max(valid_scores_last_step)
                
        best_last_step_ddg = -best_last_step_score
        # The best overall score and step are now returned directly
        best_overall_ddg = -best_overall_score
        
        # final_score = scoring_function(generated_protein)
        # final_ddg = -final_score
        
        # final_log_entry = (
        #     f"\n\n" + "="*35 + "\n    FINAL OPTIMIZED RESULT \n" + "="*35 + "\n"
        #     f"Optimized Sequence:\n{generated_protein.sequence}\n\n"
        #     f"Final Score (=-ΔΔG): {final_score:.4f}\n"
        #     f"Final FoldX ΔΔG vs. Wild-Type: {final_ddg:.4f} kcal/mol\n" + "="*35 + "\n"
        # )

        final_log_entry = (
            f"\n\n{'='*35}\n   FINAL RESULTS SUMMARY\n{'='*35}\n"
            f"\n--- Best from Final Step (Step {actual_last_step}) ---\n"
            f"Score (=-ΔΔG): {best_last_step_score:.4f}\n"
            f"FoldX ΔΔG vs. Wild-Type: {best_last_step_ddg:.4f} kcal/mol\n"
            f"\n--- Best Overall Result ---\n"
            f"Found in Step: {best_overall_step}\n"
            f"Optimized Sequence:\n{generated_protein.sequence}\n\n"
            f"Final Score (=-ΔΔG): {best_overall_score:.4f}\n"
            f"Final FoldX ΔΔG vs. Wild-Type: {best_overall_ddg:.4f} kcal/mol\n"
            f"{'='*35}\n"
        )
        
        with open(log_filepath, 'a') as f: f.write(final_log_entry)
        print(final_log_entry)
        results_dir = "../../results"
        os.makedirs(results_dir, exist_ok=True)
        plot_filename = f"ddg_history_{os.path.splitext(PDB_FILENAME)[0]}_{timestamp}.png"
        # plot_filename = f"ddg_history_{timestamp}.png"
        plot_filepath = os.path.join(results_dir, plot_filename)
        plot_ddg_history(all_scores, save_path=plot_filepath)


if __name__ == "__main__":
    main()

    


import os
import uuid
import shutil
import subprocess
import hashlib
import random
import time
from typing import List, Tuple, Optional
from typing import Optional, Tuple, Dict
import torch
from tqdm import tqdm
from esm.models.esm3 import ESM3
from esm.sdk.api import ESMProtein, GenerationConfig
from guided_generation import ESM3GuidedDecoding, GuidedDecodingScoringFunction
from multiprocessing import Pool, cpu_count, Manager
from functools import partial
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from multiprocessing import set_start_method
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)


AA3_TO_AA1 = {"ALA":"A","ARG":"R","ASN":"N","ASP":"D","CYS":"C","GLN":"Q","GLU":"E","GLY":"G","HIS":"H","ILE":"I","LEU":"L","LYS":"K","MET":"M","PHE":"F","PRO":"P","SER":"S","THR":"T","TRP":"W","TYR":"Y","VAL":"V"}

def parse_pdb_chain_sequence_with_mapping(pdb_path: str, chain_id: str) -> Tuple[str, List[str]]:
    seq, mapping, seen = [], [], set()
    with open(pdb_path, "r") as fh:
        for line in fh:
            if not line.startswith("ATOM") or line[21].strip() != chain_id: continue
            resname = line[17:20].strip().upper()
            if resname not in AA3_TO_AA1: continue
            resseq, icode = line[22:26].strip(), line[26].strip()
            key = (resseq, icode)
            if key in seen: continue
            seen.add(key)
            seq.append(AA3_TO_AA1[resname])
            mapping.append(resseq + (icode if icode else ""))
    return "".join(seq), mapping

def foldx_repair_pdb(foldx_exec: str, workdir: str, pdb_filename: str,timeout_sec: int) -> str:
    print(f"[INFO] Running FoldX RepairPDB on {pdb_filename}...")
    cmd = [foldx_exec, "--command=RepairPDB", f"--pdb={pdb_filename}"]
    proc = subprocess.run(cmd, cwd=workdir, capture_output=True, text=True, timeout=timeout_sec)
    if proc.returncode != 0: raise RuntimeError(f"RepairPDB failed. FoldX stderr:\n{proc.stderr}")
    repaired_path = os.path.join(workdir, f"{os.path.splitext(pdb_filename)[0]}_Repair.pdb")
    if not os.path.isfile(repaired_path): raise FileNotFoundError(f"RepairPDB finished but output file not found: {repaired_path}")
    print(f"[OK] PDB repaired. Output: {repaired_path}")
    return repaired_path

def get_cache_path(seq: str, cache_dir: str) -> str:
    h = hashlib.sha256(seq.encode()).hexdigest()
    return os.path.join(cache_dir, f"ddg_{h[:16]}.joblib")

def parse_ddg_from_fxout(fxout_path: str) -> Optional[float]:
    try:
        with open(fxout_path, "r") as f: lines = [line.strip() for line in f if line.strip()]
        header_line_index, header = -1, []
        for i, line in enumerate(lines):
            if "pdb" in line.lower() and ("total energy" in line.lower() or "ddg" in line.lower()):
                header_line_index, header = i, line.split('\t')
                break
        if header_line_index == -1: return None
        data_line_index = header_line_index + 1
        if len(lines) <= data_line_index: return None
        data_line, possible_colnames = lines[data_line_index].split('\t'), ["ddg", "total energy", "energydiff", "interaction energy"]
        energy_col_idx = -1
        for i, col_name in enumerate(header):
            if col_name.lower().strip().replace("_", " ") in possible_colnames:
                energy_col_idx = i
                break
        if energy_col_idx == -1 or len(data_line) <= energy_col_idx: return None
        return float(data_line[energy_col_idx])
    except Exception as e:
        print(f"[WARN] Could not parse ddG from {fxout_path}: {e}")
        return None

def get_foldx_mutation_string(candidate_seq: str, wt_seq: str, chain_id: str, seq_to_pdb_map: List[str]) -> str:
    mutations = []
    for i, (wt_aa, mut_aa) in enumerate(zip(wt_seq, candidate_seq)):
        if wt_aa != mut_aa: mutations.append(f"{wt_aa}{chain_id}{seq_to_pdb_map[i]}{mut_aa}")
    return ",".join(mutations) + ";" if mutations else ";"

# def compute_ddg_with_foldx(
#     seq: str, wt_seq: str, chain_id: str, seq_to_pdb_map: List[str],
#     repaired_pdb_path: str, foldx_exec: str, foldx_workdir: str, cache_dir: str,
#     number_of_runs: int, timeout_sec: int, cleanup_tmp: bool, verbose_foldx: bool
# ) -> Optional[float]:
#     import joblib
#     # MODIFIED: Passes cache_dir to get_cache_path
#     cache_path = get_cache_path(seq, cache_dir)
#     if os.path.exists(cache_path): return joblib.load(cache_path)
    
#     mutation_str = get_foldx_mutation_string(seq, wt_seq, chain_id, seq_to_pdb_map)
#     if not mutation_str or mutation_str == ";": return 0.0
    
#     tmp_dir = os.path.join(foldx_workdir, f"foldx_run_{uuid.uuid4().hex[:8]}")
#     os.makedirs(tmp_dir)
#     ddg = None
#     try:
#         repaired_base = os.path.basename(repaired_pdb_path)
#         shutil.copy2(repaired_pdb_path, os.path.join(tmp_dir, repaired_base))
#         shutil.copy2(os.path.join(foldx_workdir, "rotabase.txt"), tmp_dir)
#         with open(os.path.join(tmp_dir, "individual_list.txt"), "w") as f: f.write(mutation_str + "\n")
#         cmd = [foldx_exec, "--command=BuildModel", f"--pdb={repaired_base}", "--mutant-file=individual_list.txt", f"--numberOfRuns={number_of_runs}"]
#         proc = subprocess.run(cmd, cwd=tmp_dir, capture_output=True, text=True, timeout=timeout_sec)
#         if verbose_foldx: print(f"\n--- FOLDX Log for {seq[:10]}... ---\nSTDOUT:\n{proc.stdout}\nSTDERR:\n{proc.stderr}\n---------------------\n")
#         if proc.returncode == 0:
#             avg_fxout = os.path.join(tmp_dir, f"Average_{os.path.splitext(repaired_base)[0]}.fxout")
#             ddg = parse_ddg_from_fxout(avg_fxout)
#     except Exception: pass
#     finally:
#         if cleanup_tmp: shutil.rmtree(tmp_dir, ignore_errors=True)
#     if ddg is not None: joblib.dump(ddg, cache_path)
#     return ddg

# for reporting the foldx output to log

def compute_ddg_with_foldx(
    seq: str, wt_seq: str, chain_id: str, seq_to_pdb_map: List[str],
    repaired_pdb_path: str, foldx_exec: str, foldx_workdir: str, cache_dir: str,
    number_of_runs: int, timeout_sec: int, cleanup_tmp: bool, verbose_foldx: bool,
    step: int, log_file_path: str  
) -> Optional[float]:
    import joblib
    cache_path = get_cache_path(seq, cache_dir)
    if os.path.exists(cache_path): return joblib.load(cache_path)
    
    mutation_str = get_foldx_mutation_string(seq, wt_seq, chain_id, seq_to_pdb_map)
    if not mutation_str or mutation_str == ";": return 0.0
    
    tmp_dir = os.path.join(foldx_workdir, f"foldx_run_{uuid.uuid4().hex[:8]}")
    os.makedirs(tmp_dir)
    ddg = None
    try:
        repaired_base = os.path.basename(repaired_pdb_path)
        shutil.copy2(repaired_pdb_path, os.path.join(tmp_dir, repaired_base))
        shutil.copy2(os.path.join(foldx_workdir, "rotabase.txt"), tmp_dir)
        with open(os.path.join(tmp_dir, "individual_list.txt"), "w") as f: f.write(mutation_str + "\n")
        cmd = [foldx_exec, "--command=BuildModel", f"--pdb={repaired_base}", "--mutant-file=individual_list.txt", f"--numberOfRuns={number_of_runs}"]
        
        proc = subprocess.run(cmd, cwd=tmp_dir, capture_output=True, text=True, timeout=timeout_sec)

        if step == 1 and log_file_path:
            with open(log_file_path, 'a') as f:
                f.write(f"\n--- FOLDX Log for sequence {seq[:15]}... ---\n")
                f.write(f"STDOUT:\n{proc.stdout}\n")
                f.write(f"STDERR:\n{proc.stderr}\n")
                f.write("---------------------------------------------------\n")
        
        if proc.returncode == 0:
            avg_fxout = os.path.join(tmp_dir, f"Average_{os.path.splitext(repaired_base)[0]}.fxout")
            ddg = parse_ddg_from_fxout(avg_fxout)
    except Exception as e:
        if log_file_path:
            with open(log_file_path, 'a') as f:
                f.write(f"[ERROR] Exception during FoldX run for seq {seq[:15]}: {e}\n")
    finally:
        if cleanup_tmp: shutil.rmtree(tmp_dir, ignore_errors=True)
    
    if ddg is not None: joblib.dump(ddg, cache_path)
    return ddg

# class FoldXScorer(GuidedDecodingScoringFunction):
#     # MODIFIED: __init__ now accepts **kwargs to catch all extra settings
#     def __init__(self, wt_seq, chain_id, seq_to_pdb_map, repaired_pdb_path, **kwargs):
#         self.wt_seq = wt_seq
#         self.chain_id = chain_id
#         self.seq_to_pdb_map = seq_to_pdb_map
#         self.repaired_pdb_path = repaired_pdb_path
#         self.kwargs = kwargs # Store all other foldx settings
    
#     # MODIFIED: __call__ now passes the stored settings along to the compute function
#     def __call__(self, protein: ESMProtein) -> float:
#         ddg = compute_ddg_with_foldx(
#             protein.sequence, self.wt_seq, self.chain_id, self.seq_to_pdb_map, self.repaired_pdb_path, **self.kwargs
#         )
#         if ddg is None: return float("-inf")
#         score = -float(ddg)
#         return score



# class FoldXScorer(GuidedDecodingScoringFunction):
#     def __init__(self, wt_seq, chain_id, seq_to_pdb_map, repaired_pdb_path, **kwargs):
#         self.wt_seq = wt_seq
#         self.chain_id = chain_id
#         self.seq_to_pdb_map = seq_to_pdb_map
#         self.repaired_pdb_path = repaired_pdb_path
#         self.kwargs = kwargs
#         # NEW: Define the set of invalid characters
#         self.invalid_amino_acids = {'B', 'J', 'O', 'U', 'X', 'Z'}
    
#     def __call__(self, protein: ESMProtein) -> float:
#         sequence = protein.sequence

#         # MODIFIED: Expanded check for any invalid character
#         if any(char in self.invalid_amino_acids for char in sequence):
#             print(f"[SCORE] Invalid sequence with non-standard amino acid found. Score: -inf")
#             return float("-inf")

#         # The rest of the function is the same
#         ddg = compute_ddg_with_foldx(
#             sequence, self.wt_seq, self.chain_id, self.seq_to_pdb_map, self.repaired_pdb_path, **self.kwargs
#         )
#         if ddg is None: 
#             return float("-inf")
            
#         score = -float(ddg)
#         return score

# for reporting the foldx output to log

class FoldXScorer(GuidedDecodingScoringFunction):
    def __init__(self, wt_seq, chain_id, seq_to_pdb_map, repaired_pdb_path, **kwargs):
        self.wt_seq = wt_seq
        self.chain_id = chain_id
        self.seq_to_pdb_map = seq_to_pdb_map
        self.repaired_pdb_path = repaired_pdb_path
        self.kwargs = kwargs
        self.invalid_amino_acids = {'B', 'J', 'O', 'U', 'X', 'Z'}

    def __call__(self, protein: ESMProtein, step: int, log_file_path: str) -> float:
        sequence = protein.sequence

        if any(char in self.invalid_amino_acids for char in sequence):
            print(f"[SCORE] Invalid sequence with non-standard amino acid found. Score: -inf")
            return float("-inf")

        ddg = compute_ddg_with_foldx(
            sequence, self.wt_seq, self.chain_id, self.seq_to_pdb_map, 
            self.repaired_pdb_path, step=step, log_file_path=log_file_path, **self.kwargs
        )
        
        if ddg is None: 
            return float("-inf")
            
        score = -float(ddg)
        return score



def plot_ddg_history(all_scores_history: Dict, save_path: str):
    """
    Creates and saves a box plot of ddG scores over generation steps,
    with individual data points overlaid.
    """
    plot_data = []
    for step, scores in all_scores_history.items():
        for score in scores:
            if score is not None and score != float('-inf'):
                ddg = -score
                plot_data.append({'step': step, 'ddg': ddg})
    
    if not plot_data:
        print("[WARN] No valid scores to plot.")
        return

    df = pd.DataFrame(plot_data)
    
    plt.style.use('seaborn-v0_8-whitegrid')
    fig, ax = plt.subplots(figsize=(12, 7))
    
    sns.boxplot(x='step', y='ddg', data=df, ax=ax, palette="viridis")
    
    sns.stripplot(
        x='step', 
        y='ddg', 
        data=df, 
        ax=ax, 
        color='black', 
        alpha=0.6, 
        jitter=True  
    )
    
    ax.axhline(0, color='red', linestyle='--', linewidth=2, label='Wild-Type Stability (ΔΔG = 0)')
    ax.set_title('Distribution of Predicted ΔΔG Scores per Generation Step', fontsize=16)
    ax.set_xlabel('Generation Step', fontsize=12)
    ax.set_ylabel('Predicted ΔΔG (kcal/mol)', fontsize=12)
    ax.legend()
    
    plt.savefig(save_path, dpi=300, bbox_inches='tight')
    print(f"\n[INFO] Plot saved to: {save_path}")
    plt.show()


from abc import ABC, abstractmethod
import os
import attr
import torch
import time
from tqdm import tqdm
from typing import Optional, Tuple, Dict
# NEW: Import Pool and cpu_count for parallel processing
from multiprocessing import Pool, cpu_count

from esm.models.esm3 import ESM3
from esm.sdk.api import (
    ESM3InferenceClient,
    ESMProtein,
    ESMProteinError,
    ESMProteinTensor,
    SamplingConfig,
    SamplingTrackConfig,
)
from esm.sdk.forge import ESM3ForgeInferenceClient
from esm.tokenization import get_esm3_model_tokenizers
from functools import partial

class GuidedDecodingScoringFunction(ABC):
    @abstractmethod
    def __call__(self, protein: ESMProtein) -> float:
        pass


class ESM3GuidedDecoding:
    """This class can be used to perform derivative-free guided decoding..."""

    def __init__(
        self,
        client: ESM3InferenceClient,
        scoring_function: GuidedDecodingScoringFunction,
    ):
        if isinstance(client, ESM3):
            self.tokenizers = client.tokenizers
        elif isinstance(client, ESM3ForgeInferenceClient):
            self.tokenizers = get_esm3_model_tokenizers(client.model)
        else:
            raise ValueError(
                "client must be an instance of ESM3 or ESM3ForgeInferenceClient"
            )

        self.client = client
        self.scoring_function = scoring_function

    def guided_generate(
        self,
        protein: ESMProtein,
        num_decoding_steps: int,
        num_samples_per_step: int,
        denoised_prediction_temperature: float = 0.0,
        track: str = "sequence",
        verbose: bool = True,
        num_workers: int = 1,
        log_file_path: Optional[str] = None
    ) -> Tuple[ESMProtein, Dict]:
        print(f"\n[START] Initial input ESMProtein:\n  sequence = {protein.sequence[:100]}...\n")
        protein_tensor = self.client.encode(protein)
        assert not isinstance(protein_tensor, ESMProteinError)

        if verbose:
            pbar = tqdm(range(num_decoding_steps), desc="Guided Generation")
        else:
            pbar = range(num_decoding_steps)

        best_overall_score = -float('inf')
        best_overall_tensor = protein_tensor
        best_overall_sequence = protein.sequence 
        best_overall_step = 0                   

        all_scores_history = {}

        for step in pbar:
            step_total_start_time = time.time()
            current_masked_positions = self.get_number_of_masked_positions(protein_tensor, track=track)
            if current_masked_positions == 0:
                print("\n[INFO] No masked positions remaining. Finishing generation.")
                break
            
            # --- MODIFIED LOGIC: PROPORTIONAL UNMASKING SCHEDULE ---
            remaining_steps = num_decoding_steps - step
            num_to_unmask = max(1, current_masked_positions // remaining_steps) if remaining_steps > 0 else current_masked_positions
            
            
            print(f"\n[STEP {step + 1}/{num_decoding_steps}] Unmasking {num_to_unmask} of {current_masked_positions} positions...")

            candidate_gen_start_time = time.time()
            candidate_tensors = [self.randomly_unmask_positions(protein_tensor, num_to_unmask, track=track) for _ in range(num_samples_per_step)]
            
            if log_file_path:
                with open(log_file_path, 'a') as f:
                    f.write(f"\n--- Step {step + 1} Partial Sequences ---\n")
                    for i, tensor in enumerate(candidate_tensors):
                        partial_protein = self.client.decode(tensor)
                        f.write(f"[Candidate {i+1}] Partial Sequence: {partial_protein.sequence}\n")

            denoised_proteins = [self.predict_denoised(tensor, temperature=denoised_prediction_temperature) for tensor in candidate_tensors]
                   
            candidate_gen_duration = time.time() - candidate_gen_start_time
            print(f"[INFO] Candidate generation (GPU) finished in {candidate_gen_duration:.2f} seconds.")
        

            print(f"[INFO] Generated {len(denoised_proteins)} candidates for scoring.")

            print(f"[INFO] Scoring candidates in parallel using {num_workers} workers...")
            # step_start_time = time.time(). #previous timer
            scoring_start_time = time.time()

             # MODIFIED: Use partial to pass step and log_file_path to the scorer
            scorer_with_context = partial(
                self.scoring_function, 
                step=step + 1, 
                log_file_path=log_file_path
            )

            with Pool(processes=num_workers) as pool:
                # scores = pool.map(self.scoring_function, denoised_proteins)
                scores = pool.map(scorer_with_context, denoised_proteins)
            # step_duration = time.time() - step_start_time
            scoring_duration = time.time() - scoring_start_time
            # print(f"[INFO] Step scoring finished in {step_duration:.2f} seconds.")
            print(f"[INFO] Step scoring finished in {scoring_duration:.2f} seconds.")


            all_scores_history[step + 1] = scores

            best_score_in_step = -float('inf')
            best_tensor_in_step = None
            best_sequence_in_step = ""  #added

            if log_file_path:
                with open(log_file_path, 'a') as f:
                    f.write(f"\n--- Step {step + 1} Denoised Results ---\n")
                    for i, (p, score) in enumerate(zip(denoised_proteins, scores)):
                        ddg = -score
                        ptm_str = f"pTM: {p.ptm.item():.4f}" if hasattr(p, 'ptm') and p.ptm is not None else "pTM: N/A"
                        f.write(f"\n[Candidate {i+1}]\n")
                        f.write(f"  Denoised Sequence: {p.sequence}\n")
                        f.write(f"  {ptm_str}\n")
                        f.write(f"  Score: {score:.4f} | ddG: {ddg:.3f} kcal/mol\n")
                        if score > best_score_in_step:
                            best_score_in_step = score
                            best_tensor_in_step = candidate_tensors[i]
                            best_sequence_in_step = p.sequence #added
                            
            else:
                for i, score in enumerate(scores):
                    if score > best_score_in_step:
                        best_score_in_step = score
                        best_tensor_in_step = candidate_tensors[i]
                        
            
            if best_score_in_step > best_overall_score:
                best_overall_score = best_score_in_step
                best_overall_tensor = best_tensor_in_step
                best_overall_sequence = best_sequence_in_step 
                best_overall_step = step + 1                 
            
            protein_tensor = best_tensor_in_step
            step_total_duration = time.time() - step_total_start_time

            with open(log_file_path, 'a') as f:
                f.write("\n" + "-"*25 + f" Step {step + 1} Summary " + "-"*25 + "\n")
                f.write(f"Best Candidate in Step:\n{best_sequence_in_step}\n")
                f.write(f"Best Score in Step (=-ΔΔG): {best_score_in_step:.4f}\n")
                f.write(f"This sequence will be used as the template for Step {step + 2}.\n")
                f.write("-" * 65 + "\n")

            print(f"\n[STEP {step + 1}] Best Candidate Score in Step: {best_score_in_step:.4f}")

            if verbose:
                pbar.set_description(f"Best score so far: {best_overall_score:.4f}")

            with open(log_file_path, 'a') as f:
                f.write(f"Timing for Step {step + 1}:\n")
                f.write(f"  - Candidate Generation (GPU): {candidate_gen_duration:.2f} s\n")
                f.write(f"  - Scoring (CPU/FoldX):        {scoring_duration:.2f} s\n")
                f.write(f"  - Total Step Time:              {step_total_duration:.2f} s\n")
        

            # NEW TIMER: End timer for the whole step
            # step_total_duration = time.time() - step_total_start_time
            print(f"\n[INFO] Total time for Step {step + 1}: {step_total_duration:.2f} seconds.")

        print("\n[FINAL] Performing full denoising of the best candidate to ensure completion...")
        final_complete_protein = self.predict_denoised(best_overall_tensor, temperature=0.0)
        assert not isinstance(final_complete_protein, ESMProteinError)
        
        # return final_complete_protein,all_scores_history

        return final_complete_protein, all_scores_history, best_overall_score, best_overall_step




    def reward_function(
        self,
        protein_tensor: ESMProteinTensor,
        denoised_prediction_temperature: float = 0.0,
    ) -> float:
        denoised_protein = self.predict_denoised(
            protein_tensor, temperature=denoised_prediction_temperature
        )
        print(f"[Denoised] Sequence: {denoised_protein.sequence[:60]}...")
        if hasattr(denoised_protein, "ptm"):
            print(f"[Denoised] pTM: {float(denoised_protein.ptm):.4f}")
        return self.scoring_function(denoised_protein)

    def get_number_of_masked_positions(
        self, protein_tensor: ESMProteinTensor, track: str = "sequence"
    ) -> int:
        assert isinstance(protein_tensor, ESMProteinTensor)
        track_tensor = getattr(protein_tensor, track)
        track_tokenizer = getattr(self.tokenizers, track)
        is_mask = track_tensor == track_tokenizer.mask_token_id
        return is_mask.sum().item()  # type: ignore

    def randomly_unmask_positions(
        self,
        protein_tensor: ESMProteinTensor,
        num_positions_to_unmask: int,
        temperature: float = 1.0,
        track: str = "sequence",
    ) -> ESMProteinTensor:
        track_tensor = getattr(protein_tensor, track)
        assert track_tensor is not None
        protein_tensor = attr.evolve(protein_tensor)
        setattr(protein_tensor, track, track_tensor.clone())

        track_tensor = getattr(protein_tensor, track)
        track_tokenizer = getattr(self.tokenizers, track)

        is_mask = track_tensor == track_tokenizer.mask_token_id
        num_masked_positions = is_mask.sum().item()

        if num_positions_to_unmask > num_masked_positions:
            num_positions_to_unmask = num_masked_positions  # type: ignore

        mask_indices = is_mask.nonzero(as_tuple=False)
        mask_indices = mask_indices[torch.randperm(mask_indices.size(0))]
        mask_indices = mask_indices[:num_positions_to_unmask]

        sampling_config = SamplingConfig()
        setattr(sampling_config, track, SamplingTrackConfig(temperature=temperature))

        denoised_protein_tensor_output = self.client.forward_and_sample(
            protein_tensor, sampling_configuration=sampling_config
        )
        assert not isinstance(denoised_protein_tensor_output, ESMProteinError)
        denoised_protein_tensor = denoised_protein_tensor_output.protein_tensor
        output_track_tensor = getattr(denoised_protein_tensor, track)
        assert output_track_tensor is not None
        track_tensor[mask_indices] = output_track_tensor[mask_indices]
        setattr(protein_tensor, track, track_tensor)

        return protein_tensor

    def predict_denoised(
        self, protein_tensor: ESMProteinTensor, temperature: float = 0.0
    ) -> ESMProtein:
        denoised_protein_tensor_output = self.client.forward_and_sample(
            protein_tensor,
            sampling_configuration=SamplingConfig(
                sequence=SamplingTrackConfig(temperature=temperature),
                structure=SamplingTrackConfig(temperature=temperature),
            ),
        )
        assert not isinstance(denoised_protein_tensor_output, ESMProteinError)
        denoised_protein_tensor = denoised_protein_tensor_output.protein_tensor
        denoised_protein = self.client.decode(denoised_protein_tensor)
        assert not isinstance(denoised_protein, ESMProteinError)
        return denoised_protein

    def maybe_add_default_structure_tokens(
        self, protein_tensor: ESMProteinTensor
    ) -> ESMProteinTensor:
        empty_protein_tensor = ESMProteinTensor.empty(
            len(protein_tensor) - 2,
            tokenizers=self.tokenizers,
            device=protein_tensor.device,
        )
        if protein_tensor.structure is None:
            setattr(protein_tensor, "structure", empty_protein_tensor.structure)
        else:
            print("Warning: structure already exists in protein_tensor")
        return protein_tensor



# In setup.py

from setuptools import setup, find_packages

setup(
    name="esm_foldx_guidedgeneration",
    version="0.1.0",
    author="Amitash Nanda",
    author_email="amitashnanda01@gmail.com",
    description="A tool for Guided Generation based Protein Design and Engineering using FoldX, or any other properties adding to custom-scoring function.",
    long_description=open('README.md').read(), 
    long_description_content_type="text/markdown",
    url="https://github.com/amitashnanda/ESM3-Guided-Generation-Based-Protein-Engineering.git", 
    
    package_dir={"": "src"}, 
    packages=find_packages(where="src"), 
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'torch',
        'pandas',
        'seaborn',
        'matplotlib',
        'esm', 
        'biopython',
        'huggingface-hub',
        'tqdm'
    ],

    entry_points={
        'console_scripts': [
            'esm_foldx_guidedgeneration = esm_foldx_guidedgeneration.main:main',
        ],
    },
)